from sindripy.value_objects.decorators.validation import validate
from sindripy.value_objects.errors.incorrect_value_type_error import IncorrectValueTypeError
from sindripy.value_objects.errors.required_value_error import RequiredValueError
from sindripy.value_objects.value_object import ValueObject


class Boolean(ValueObject[bool]):
    """
    A value object that wraps boolean values with validation.

    This class provides a base implementation for creating value objects that
    represent boolean values in the domain. It ensures that the wrapped value
    is a valid boolean and not None.

    The class includes built-in validation for:
    - Required value (not None)
    - Type checking (must be a boolean)

    Inherits all functionality from ValueObject including immutability,
    equality comparison, string representation, and hashing.

    Example:
        ```python
        class IsActive(Boolean):
             @validate
             def _validate_true_for_premium(self, value: bool) -> None:
                 # Custom business logic can be added here
                 pass

        is_active = IsActive(True)
        is_active.value  # True
        str(is_active)  # 'True'
        ```
    """

    @validate
    def _ensure_has_value(self, value: bool) -> None:
        if value is None:
            raise RequiredValueError

    @validate
    def _ensure_value_is_boolean(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise IncorrectValueTypeError(value, bool)



"""Top level package for the value-crafter library.

This module exposes the most common entry points so the public API is
available through the ``sindripy`` namespace when the library is
installed as a dependency.
"""

from sindripy import mothers, value_objects

__all__ = ["mothers", "value_objects"]
__version__ = "0.1.6"



from sindripy.mothers.object_mother import ObjectMother


class ListPrimitivesMother(ObjectMother):
    """Generate list primitive values for testing."""

    @staticmethod
    def empty() -> list:
        """Generate an empty list."""
        return []



class SindriValidationError(Exception):
    """Base class for all controlled errors during validation of value objects."""

    def __init__(self, message: str) -> None:
        self._message = message
        super().__init__(self._message)

    @property
    def message(self) -> str:
        return self._message



from sindripy.value_objects.errors.sindri_validation_error import SindriValidationError


class InvalidIdFormatError(SindriValidationError):
    def __init__(self) -> None:
        super().__init__(
            message="Id must be a valid UUID",
        )



"""Public facade for value object implementations.

This module re-exports the most common value objects so they can be
imported directly from :mod:`sindripy.value_object`.
"""

from sindripy.value_objects.aggregate import Aggregate
from sindripy.value_objects.decorators.validation import validate
from sindripy.value_objects.errors.sindri_validation_error import SindriValidationError
from sindripy.value_objects.identifiers.string_uuid import StringUuid
from sindripy.value_objects.primitives.boolean import Boolean
from sindripy.value_objects.primitives.float import Float
from sindripy.value_objects.primitives.integer import Integer
from sindripy.value_objects.primitives.list import List
from sindripy.value_objects.primitives.string import String
from sindripy.value_objects.value_object import ValueObject

__all__ = [
    "Aggregate",
    "validate",
    "StringUuid",
    "Boolean",
    "Float",
    "Integer",
    "List",
    "String",
    "ValueObject",
    "SindriValidationError",
]



from collections.abc import Iterator
from typing import Any, Generic, TypeVar, get_args, get_origin

from sindripy._compat import Self, override
from sindripy.value_objects.decorators.validation import validate
from sindripy.value_objects.errors.incorrect_value_type_error import IncorrectValueTypeError
from sindripy.value_objects.errors.required_value_error import RequiredValueError
from sindripy.value_objects.value_object import ValueObject

T = TypeVar("T")


class List(ValueObject[list[T]], Generic[T]):
    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """
        Initialize subclass with proper type parameter validation.

        This method ensures that any subclass of List is properly parameterized
        with a type argument and extracts the element type for validation purposes. It
        is run automatically when a subclass is created, at definition time.
        """
        super().__init_subclass__(**kwargs)

        cls._validate_class_has_type_parameters()
        list_base = cls._find_parameterized_list_base()
        element_type = cls._extract_element_type_from_base(list_base)
        cls._validate_and_store_element_type(element_type)

    @classmethod
    def from_primitives(cls, value: list[Any]) -> Self:
        elements = []

        for primitive in value:
            if cls._element_is_an_aggregate_instance():
                elements.append(cls._element_type.from_primitives(primitive))
            elif cls._element_is_a_value_object_instance():
                elements.append(cls._element_type(primitive))
            else:
                elements.append(primitive)
        return cls(elements)

    @validate
    def _ensure_has_value(self, value: list[T]) -> None:
        if value is None:
            raise RequiredValueError

    @validate
    def _ensure_is_list(self, value: list[T]) -> None:
        if not isinstance(value, list):
            raise IncorrectValueTypeError(value, type[Any])

    @validate
    def _ensure_list_elements_have_expected_type(self, value: list[T]) -> None:
        cls = self.__class__

        if not hasattr(cls, "_element_type"):
            return

        element_type = cls._element_type

        if isinstance(element_type, TypeVar):
            return

        if not isinstance(element_type, type):
            return

        if cls._element_is_a_value_object_instance() or cls._element_is_a_primitive_type():
            for item in value:
                if not isinstance(item, element_type):
                    raise IncorrectValueTypeError(item, list)

    def __contains__(self, item: Any) -> bool:
        """
        Check if an item is present in the list.

        Args:
            item: The item to check for membership in the list.

        Returns:
            True if the item is in the list, False otherwise.

        Example:
            ```python
            numbers = IntList([1, 2, 3, 4, 5])
            3 in numbers  # True
            6 in numbers  # False
            ```
        """
        return item in self._value

    def __iter__(self) -> Iterator[T]:
        """
        Return an iterator over the list elements.

        Returns:
            An iterator that yields each element in the list.

        Example:
            ```python
            numbers = IntList([1, 2, 3])
            list(numbers)  # [1, 2, 3]
            for num in numbers:
                print(num)
            # 1
            # 2
            # 3
            ```
        """
        return iter(self._value)

    def __len__(self) -> int:
        """
        Return the number of elements in the list.

        Returns:
            The length of the wrapped list.

        Example:
            ```python
            numbers = IntList([1, 2, 3, 4, 5])
            len(numbers)  # 5
            ```
        """
        return len(self._value)

    def __reversed__(self) -> Iterator[T]:
        """
        Return a reverse iterator over the list elements.

        Returns:
            A reverse iterator that yields elements from the end to the beginning.

        Example:
            ```python
            numbers = IntList([1, 2, 3])
            list(reversed(numbers))  # [3, 2, 1]
            ```
        """
        return reversed(self._value)

    @override
    def __hash__(self) -> int:
        """
        Return the hash value of this value object.

        Since lists are unhashable in Python, we convert the list to a tuple
        for hashing purposes while maintaining the original list value.

        Returns:
            The hash value based on the tuple representation of the list.
        """
        return hash(tuple(self._value))

    @override
    def __eq__(self, other: Any) -> bool:
        """
        Check equality with another List value object.

        Args:
            other: The other object to compare against.

        Returns:
            True if both are List instances with the same element type and values, False otherwise.

        Example:
            ```python
            list1 = IntList([1, 2, 3])
            list2 = IntList([1, 2, 3])
            list3 = IntList([4, 5, 6])
            list1 == list2  # True
            list1 == list3  # False
            ```
        """
        if not isinstance(other, List):
            return False

        if self._element_type != other._element_type:
            return False

        return self._value == other._value

    @override
    def __repr__(self) -> str:
        """
        Return a string representation suitable for debugging.

        Returns:
            A string showing the class name and the wrapped list value.

        Example:
            ```python
            numbers = IntList([1, 2, 3])
            repr(numbers)  # 'IntList(_value=[1, 2, 3])'
            ```
        """
        return f"{self.__class__.__name__}(_value={self._value!r})"

    @override
    def __str__(self) -> str:
        """
        Return a human-readable string representation.

        Returns:
            The string representation of the wrapped list.

        Example:
            ```python
            numbers = IntList([1, 2, 3])
            str(numbers)  # '[1, 2, 3]'
            ```
        """
        return str(self._value)

    @classmethod
    def _validate_class_has_type_parameters(cls) -> None:
        """
        Validate that the class has type parameters defined.

        Raises:
            TypeError: If the class doesn't have __orig_bases__ or it's empty.
        """
        if not hasattr(cls, "__orig_bases__") or not getattr(cls, "__orig_bases__", None):
            raise TypeError(f"Class {cls.__name__} must be parameterized with a type argument")

    @classmethod
    def _find_parameterized_list_base(cls) -> Any:
        """
        Find the parameterized List base class in the inheritance chain.

        Returns:
            The parameterized List base class.

        Raises:
            TypeError: If no parameterized List base is found.
        """
        orig_bases = getattr(cls, "__orig_bases__", ())
        for base in orig_bases:
            if get_origin(base) is List:
                return base

        raise TypeError(f"Class {cls.__name__} must inherit from List[T] with a type parameter")

    @classmethod
    def _extract_element_type_from_base(cls, list_base: Any) -> Any:
        """
        Extract the element type from the parameterized List base.

        Args:
            list_base: The parameterized List base class.

        Returns:
            The element type parameter.
        """
        element_type, *_ = get_args(list_base)
        return element_type

    @classmethod
    def _validate_and_store_element_type(cls, element_type: Any) -> None:
        """
        Validate the element type and store it in the class.

        This method handles different types of type parameters:
        - TypeVar instances for generic types
        - Concrete types like int, str, etc.
        - Generic types like list[int], dict[str, int], etc.

        Args:
            element_type: The element type to validate and store.

        Raises:
            TypeError: If the element type is not a valid type parameter.
        """
        # Handle TypeVar cases: If the type is a generic type, store it directly
        if isinstance(element_type, TypeVar):
            cls._element_type = element_type
            return

        # Validate concrete types: Ensure the type parameter is actually a valid type
        if isinstance(element_type, type):
            cls._element_type = element_type
            return

        # Handle generic types like list[int], dict[str, int], etc.
        if hasattr(element_type, "__origin__") or get_origin(element_type) is not None:
            cls._element_type = element_type
            return

        raise TypeError(f"Type parameter must be a valid type, not a primitive value: {element_type}")

    @classmethod
    def _element_is_an_aggregate_instance(cls) -> bool:
        return hasattr(cls._element_type, "from_primitives")

    @classmethod
    def _element_is_a_value_object_instance(cls) -> bool:
        try:
            return isinstance(cls._element_type, type) and issubclass(cls._element_type, ValueObject)
        except TypeError:
            return False

    @classmethod
    def _element_is_a_primitive_type(cls) -> bool:
        try:
            return isinstance(cls._element_type, type) and not issubclass(cls._element_type, ValueObject)
        except TypeError:
            return False



from abc import ABC, abstractmethod
from enum import Enum
from inspect import Parameter, _empty, signature
from typing import Any

from sindripy._compat import Self, override
from sindripy.value_objects.value_object import ValueObject


class Aggregate(ABC):
    """
    Abstract base class for implementing aggregates in domain-driven design.

    Aggregates are clusters of domain objects that can be treated as a single unit.
    They ensure consistency boundaries and encapsulate business rules across
    related entities and value objects.

    This class provides utilities for:
        - Converting aggregates to/from primitive dictionaries
        - Comparing aggregates for equality
        - String representation for debugging
        - Handling nested value objects and other aggregates

    Example:
        >>> class User(Aggregate):
        ...     def __init__(self, user_id: int, name: str, email: str):
        ...         self.user_id = user_id
        ...         self.name = name
        ...         self.email = email
        ...
        >>> user = User(1, "John Doe", "john@example.com")
        >>> user.to_primitives()
        {'user_id': 1, 'name': 'John Doe', 'email': 'john@example.com'}
    """

    @abstractmethod
    def __init__(self) -> None:
        """
        Initialize the aggregate.

        This method must be implemented by all concrete aggregate classes.
        It should set up the aggregate's state and ensure all invariants are met.

        Raises:
            NotImplementedError: Always raised as this is an abstract method.

        Example:
            >>> class Product(Aggregate):
            ...     def __init__(self, product_id: str, name: str, price: float):
            ...         self.product_id = product_id
            ...         self.name = name
            ...         self.price = price
            ...         if price < 0:
            ...             raise ValueError("Price cannot be negative")
        """
        raise NotImplementedError

    @override
    def __repr__(self) -> str:
        """
        Return a string representation suitable for debugging.

        Creates a string representation showing all non-private attributes
        in a constructor-like format.

        Returns:
            A string in the format "ClassName(attr1=value1, attr2=value2, ...)"

        Example:
            >>> class Order(Aggregate):
            ...     def __init__(self, order_id: str, customer_id: int, total: float):
            ...         self.order_id = order_id
            ...         self.customer_id = customer_id
            ...         self.total = total
            ...
            >>> order = Order("ORD-001", 123, 99.99)
            >>> repr(order)
            "Order(customer_id=123, order_id='ORD-001', total=99.99)"
        """
        attributes = []
        for key, value in self._to_dict().items():
            attributes.append(f"{key}={value!r}")

        return f"{self.__class__.__name__}({', '.join(attributes)})"

    @override
    def __eq__(self, other: Self) -> bool:
        """
        Check equality with another aggregate of the same type.

        Two aggregates are considered equal if they are of the same class
        and all their non-private attributes have equal values.

        Args:
            other: Another aggregate of the same type to compare with.

        Returns:
            True if both aggregates have equal attributes, False otherwise.
            NotImplemented if comparing with a different type.

        Example:
            >>> class Category(Aggregate):
            ...     def __init__(self, cat_id: int, name: str):
            ...         self.cat_id = cat_id
            ...         self.name = name
            ...
            >>> cat1 = Category(1, "Electronics")
            >>> cat2 = Category(1, "Electronics")
            >>> cat3 = Category(2, "Books")
            >>> cat1 == cat2
            True
            >>> cat1 == cat3
            False
        """
        if not isinstance(other, self.__class__):
            return NotImplemented

        return self._to_dict() == other._to_dict()

    def _to_dict(self, *, ignore_private: bool = True) -> dict[str, Any]:
        """
        Convert the aggregate to a dictionary representation.

        Extracts all instance attributes and converts them to a dictionary,
        optionally filtering out private attributes (those starting with
        double underscore and class name).

        Args:
            ignore_private: Whether to exclude private attributes from the result.
                           Defaults to True.

        Returns:
            A dictionary mapping attribute names to their values.

        Example:
            >>> class Invoice(Aggregate):
            ...     def __init__(self, invoice_id: str, amount: float):
            ...         self.invoice_id = invoice_id
            ...         self.amount = amount
            ...         self._calculated_tax = amount * 0.1  # private attribute
            ...         self.__Invoice__secret = "hidden"    # name-mangled private
            ...
            >>> invoice = Invoice("INV-001", 100.0)
            >>> invoice._to_dict()
            {'invoice_id': 'INV-001', 'amount': 100.0, 'calculated_tax': 100.0}
            >>> invoice._to_dict(ignore_private=False)
            {'invoice_id': 'INV-001', 'amount': 100.0, 'calculated_tax': 100.0, 'secret': 'hidden'}
        """
        dictionary: dict[str, Any] = {}
        for key, value in self.__dict__.items():
            if ignore_private and key.startswith(f"_{self.__class__.__name__}__"):
                continue  # ignore private attributes

            key = key.replace(f"_{self.__class__.__name__}__", "")

            if key.startswith("_"):
                key = key[1:]

            dictionary[key] = value

        return dictionary

    @classmethod
    def from_primitives(cls, primitives: dict[str, Any]) -> Self:
        """
        Create an aggregate instance from a dictionary of primitive values.

        This factory method constructs an aggregate by mapping dictionary keys
        to constructor parameters. All required constructor parameters must be
        present in the primitives dictionary.

        Args:
            primitives: A dictionary mapping parameter names to their values.
                       Must contain all required constructor parameters.

        Returns:
            A new instance of the aggregate class.

        Raises:
            TypeError: If primitives is not a dictionary with string keys.
            ValueError: If required parameters are missing or extra parameters are provided.

        Example:
            >>> class Customer(Aggregate):
            ...     def __init__(self, customer_id: int, name: str, email: str = None):
            ...         self.customer_id = customer_id
            ...         self.name = name
            ...         self.email = email
            ...
            >>> data = {"customer_id": 42, "name": "Alice Smith"}
            >>> customer = Customer.from_primitives(data)
            >>> customer.name
            'Alice Smith'
            >>> customer.customer_id
            42
            >>>
            >>> # Missing required parameter
            >>> Customer.from_primitives({"name": "Bob"})  # Raises ValueError
            >>>
            >>> # Extra parameter
            >>> Customer.from_primitives({"customer_id": 1, "name": "Charlie", "age": 30})  # Raises ValueError
        """
        if not isinstance(primitives, dict) or not all(isinstance(key, str) for key in primitives):
            raise TypeError(f'{cls.__name__} primitives <<<{primitives}>>> must be a dictionary of strings. Got <<<{type(primitives).__name__}>>> type.')  # noqa: E501  # fmt: skip

        constructor_signature = signature(obj=cls.__init__)
        parameters: dict[str, Parameter] = {parameter.name: parameter for parameter in constructor_signature.parameters.values() if parameter.name != 'self'}  # noqa: E501  # fmt: skip
        missing = {name for name, parameter in parameters.items() if parameter.default is _empty and name not in primitives}  # noqa: E501  # fmt: skip
        extra = set(primitives) - parameters.keys()

        if missing or extra:
            cls._raise_value_constructor_parameters_mismatch(primitives=set(primitives), missing=missing, extra=extra)

        return cls(**primitives)

    @classmethod
    def _raise_value_constructor_parameters_mismatch(
        cls,
        primitives: set[str],
        missing: set[str],
        extra: set[str],
    ) -> None:
        """
        Raise a detailed ValueError for constructor parameter mismatches.

        This helper method generates informative error messages when the
        primitives dictionary doesn't match the constructor signature.

        Args:
            primitives: Set of parameter names provided in the primitives dict.
            missing: Set of required parameter names that are missing.
            extra: Set of parameter names that are not in the constructor.

        Raises:
            ValueError: Always raised with detailed information about the mismatch.

        Example:
            >>> class Item(Aggregate):
            ...     def __init__(self, item_id: str, name: str, price: float):
            ...         pass
            ...
            >>> # This would trigger the error method internally:
            >>> Item.from_primitives({"item_id": "123", "description": "test"})
            Traceback (most recent call last):
                ...
            ValueError: Item primitives <<<description, item_id>>> must contain all constructor parameters. Missing parameters: <<<name, price>>> and extra parameters: <<<description>>>.
        """
        primitives_names = ", ".join(sorted(primitives))
        missing_names = ", ".join(sorted(missing))
        extra_names = ", ".join(sorted(extra))

        raise ValueError(f'{cls.__name__} primitives <<<{primitives_names}>>> must contain all constructor parameters. Missing parameters: <<<{missing_names}>>> and extra parameters: <<<{extra_names}>>>.')  # noqa: E501  # fmt: skip

    def to_primitives(self) -> dict[str, Any]:
        """
        Convert the aggregate to a dictionary of primitive values.

        Recursively converts the aggregate and all nested objects (other aggregates,
        value objects, enums) to their primitive representations. This is useful
        for serialization to JSON, database storage, or API responses.

        Returns:
            A dictionary with primitive values (strings, numbers, booleans, etc.)
            where complex objects have been converted to their primitive forms.

        Example:
            >>> from enum import Enum
            >>>
            >>> class Status(Enum):
            ...     ACTIVE = "active"
            ...     INACTIVE = "inactive"
            ...
            >>> class UserId(ValueObject[int]):
            ...     pass
            ...
            >>> class Account(Aggregate):
            ...     def __init__(self, user_id: UserId, status: Status, balance: float):
            ...         self.user_id = user_id
            ...         self.status = status
            ...         self.balance = balance
            ...
            >>> account = Account(UserId(123), Status.ACTIVE, 1500.50)
            >>> account.to_primitives()
            {'user_id': 123, 'status': 'active', 'balance': 1500.5}
            >>>
            >>> # With nested aggregates
            >>> class Order(Aggregate):
            ...     def __init__(self, account: Account, item_count: int):
            ...         self.account = account
            ...         self.item_count = item_count
            ...
            >>> order = Order(account, 3)
            >>> order.to_primitives()
            {'account': {'user_id': 123, 'status': 'active', 'balance': 1500.5}, 'item_count': 3}
        """
        primitives = self._to_dict()
        for key, value in primitives.items():
            if isinstance(value, Aggregate) or hasattr(value, "to_primitives"):
                value = value.to_primitives()

            elif isinstance(value, Enum):
                value = value.value

            elif isinstance(value, ValueObject) or hasattr(value, "value"):
                value = value.value

                if isinstance(value, Enum):
                    value = value.value

            primitives[key] = value

        return primitives



from sindripy.mothers.object_mother import ObjectMother


class FloatPrimitivesMother(ObjectMother):
    """Generate float primitive values for testing."""

    @classmethod
    def any(cls) -> float:
        """Generate any random float value."""
        return cls._faker().pyfloat()

    @classmethod
    def create(cls, is_positive: bool | None = None, min_value: float = -1000.0, max_value: float = 10000.0) -> float:
        """Generate a float value with specified constraints."""
        return cls._faker().pyfloat(positive=is_positive, min_value=min_value, max_value=max_value)

    @classmethod
    def positive(cls) -> float:
        """Generate a positive float value greater than zero."""
        return cls._faker().pyfloat(positive=True, min_value=0.1)

    @classmethod
    def negative(cls) -> float:
        """Generate a negative float value less than zero."""
        return cls._faker().pyfloat(positive=False, max_value=-0.1)

    @staticmethod
    def zero() -> float:
        """Generate zero as a float."""
        return 0.0



from typing import Any, TypeVar

from sindripy.value_objects.errors.sindri_validation_error import SindriValidationError

T = TypeVar("T")


class IncorrectValueTypeError(SindriValidationError):
    def __init__(self, value: T, expected_type: type[Any]) -> None:
        super().__init__(
            message=f"Value '{value}' is not of type {expected_type.__name__}",
        )



"""Compatibility helpers for typing features across Python versions."""

from __future__ import annotations

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

try:
    from typing import override
except ImportError:
    from typing_extensions import override

__all__ = ["Self", "override"]



from sindripy.mothers.object_mother import ObjectMother


class IntegerPrimitivesMother(ObjectMother):
    """Generate int primitive values for testing."""

    @classmethod
    def any(cls) -> int:
        """Generate any random int value."""
        return cls._faker().random_int()

    @classmethod
    def create(cls, is_positive: bool | None = None, min_value: int = -10000, max_value: int = 1000) -> int:
        """Generate an int value with specified constraints."""
        if is_positive:
            return cls._faker().random_int(min=1, max=abs(max_value))

        if is_positive is False:
            return cls._faker().random_int(min=-abs(min_value), max=-1)

        return cls._faker().random_int(min=min_value, max=max_value)

    @classmethod
    def positive(cls) -> int:
        """Generate a positive int value greater than zero."""
        return cls._faker().random_int(min=1)

    @classmethod
    def negative(cls) -> int:
        """Generate a negative int value less than zero."""
        return cls._faker().random_int(min=-(2**31), max=-1)

    @staticmethod
    def zero() -> int:
        """Generate zero as an int."""
        return 0



from sindripy.mothers.object_mother import ObjectMother


class StringPrimitivesMother(ObjectMother):
    """Generate string primitive values for testing."""

    @classmethod
    def any(cls) -> str:
        """Generate any random string value."""
        return cls._faker().word()

    @staticmethod
    def empty() -> str:
        return ""

    @classmethod
    def containing_character(cls, character: str) -> str:
        """Generate a string containing a specific character in a random position (not at beginning or end)."""
        base_word = cls._faker().word()

        if len(base_word) < 3:
            base_word = cls._faker().word() + cls._faker().word()

        character_position = cls._faker().random_int(min=1, max=len(base_word) - 1)

        return base_word[:character_position] + character + base_word[character_position + 1 :]

    @classmethod
    def ending_with(cls, character: str) -> str:
        """Generate a string ending with a specific character."""
        return cls._faker().word() + character

    @classmethod
    def beginning_with(cls, character: str) -> str:
        """Generate a string beginning with a specific character."""
        return character + cls._faker().word()

    @classmethod
    def with_length(cls, length: int) -> str:
        """Generate a string with specific length. If length is less than or equal to 0, return an empty string."""
        if length <= 0:
            return ""
        return cls._faker().pystr(min_chars=length, max_chars=length)

    @classmethod
    def text(cls) -> str:
        """Generate a text string (can contain spaces and punctuation)."""
        return cls._faker().text(max_nb_chars=200)



from uuid import UUID

from sindripy.value_objects.decorators.validation import validate
from sindripy.value_objects.errors.incorrect_value_type_error import IncorrectValueTypeError
from sindripy.value_objects.errors.invalid_id_format_error import InvalidIdFormatError
from sindripy.value_objects.errors.required_value_error import RequiredValueError
from sindripy.value_objects.value_object import ValueObject


class StringUuid(ValueObject[str]):
    """
    A value object that wraps UUID (Universally Unique Identifier) string values with validation.

    This class provides a specialized implementation for creating value objects that
    represent UUID values in the domain. It ensures that the wrapped value is a
    valid UUID string format and not None.

    The class includes built-in validation for:
    - Required value (not None)
    - Type checking (must be a string)
    - UUID format validation (must be a valid UUID format)

    Inherits all functionality from ValueObject including immutability,
    equality comparison, string representation, and hashing.

    Example:
        ```python
        class UserId(StringUuid):
            @validate
            def _validate_version(self, value: str) -> None:
                parsed_uuid = UUID(value)
                if parsed_uuid.version != 4:
                    raise ValueError("Only UUID version 4 allowed")

        user_id = UserId("123e4567-e89b-12d3-a456-426614174000")
        user_id.value  # '123e4567-e89b-12d3-a456-426614174000'
        ```
    """

    @validate
    def _ensure_has_value(self, value: str) -> None:
        if value is None:
            raise RequiredValueError

    @validate
    def _ensure_value_is_string(self, value: str) -> None:
        if not isinstance(value, str):
            raise IncorrectValueTypeError(value, str)

    @validate
    def _ensure_value_has_valid_uuid_format(self, value: str) -> None:
        try:
            UUID(value)
        except ValueError as error:
            raise InvalidIdFormatError from error



from sindripy.value_objects.decorators.validation import validate
from sindripy.value_objects.errors.incorrect_value_type_error import IncorrectValueTypeError
from sindripy.value_objects.errors.required_value_error import RequiredValueError
from sindripy.value_objects.value_object import ValueObject


class Integer(ValueObject[int]):
    """
    A value object that wraps integer values with validation.

    This class provides a base implementation for creating value objects that
    represent integer values in the domain. It ensures that the wrapped value
    is a valid integer and not None.

    The class includes built-in validation for:
    - Required value (not None)
    - Type checking (must be an integer)

    Inherits all functionality from ValueObject including immutability,
    equality comparison, string representation, and hashing.

    Example:
        ```python
        class Age(Integer):
            @validate
            def _validate_positive(self, value: int) -> None:
                if value < 0:
                    raise ValueError("Age cannot be negative")

        age = Age(25)
        age.value  # 25
        str(age)  # '25'
        ```
    """

    @validate
    def _ensure_has_value(self, value: int) -> None:
        if value is None:
            raise RequiredValueError

    @validate
    def _ensure_value_is_integer(self, value: int) -> None:
        if not isinstance(value, int):
            raise IncorrectValueTypeError(value, int)



from abc import ABC
from collections.abc import Callable
from typing import Generic, TypeVar

from sindripy._compat import Self, override

T = TypeVar("T")


class ValueObject(ABC, Generic[T]):
    """
    Abstract base class for implementing value objects with immutability and validation.

    Value objects are immutable objects that represent a descriptive aspect of the domain
    with no conceptual identity. They are equal when their values are equal.

    Type Parameters:
        T: The type of the value being wrapped by this value object.

    Attributes:
        _value: The internal value stored by this value object.

    Example:
        ```python
        from sindripy import ValueObject

        class String(ValueObject[str]):
            pass

        string = String("Hello World")
        repr(email)  # String(_value='Hello World')
        ```
    """

    __slots__ = ("_value",)
    __match_args__ = ("_value",)

    _value: T

    def __init__(self, value: T) -> None:
        """
        Initialize the value object with the given value.

        The value is validated using all methods decorated with @validate before
        being stored. Once initialized, the value cannot be modified.

        Args:
            value: The value to be wrapped by this value object.

        Raises:
            Various validation errors depending on the specific value object implementation.

        Example:
            from sindripy import ValueObject

            class String(ValueObject[str]):
                pass

            string = String("Hello World")
            repr(email)  # String(_value='Hello World')
            ```
        """
        self._validate(value)
        object.__setattr__(self, "_value", value)

    def _validate(self, value: T) -> None:
        """
        Validates the given value using all methods decorated with @validate.

        This method collects all validator methods from the class hierarchy (in reverse MRO order)
        and executes them in the order specified by their _order attribute. All validators
        must pass for the value to be considered valid.

        Args:
            value: The value to validate.

        Raises:
            Various validation errors if any validator fails.

        Example:
            ```python
            class Username(ValueObject[str]):
                @validate(order=1)
                def _validate_not_empty(self, value: str) -> None:
                    if not value.strip():
                        raise ValueError("Username cannot be empty")

                @validate(order=2)
                def _validate_length(self, value: str) -> None:
                    if len(value) < 3:
                        raise ValueError("Username must be at least 3 characters")

            username = Username("john")  # Both validators pass
            username._validate("ab")  # Would raise ValueError for length
            ```
        """
        validators: list[Callable[[T], None]] = []
        for cls in reversed(self.__class__.__mro__):
            if cls is object:
                continue

            methods: list[tuple[int, Callable[[T], None]]] = []
            for name, member in cls.__dict__.items():
                if getattr(member, "_is_validator", False):
                    validators.append(getattr(self, name))
                    order: int = getattr(member, "_order", 0)
                    methods.append((order, getattr(self, name)))

            for _, method in sorted(methods, key=lambda item: item[0]):
                validators.append(method)

        for validator in validators:
            validator(value)

    @property
    def value(self) -> T:
        """
        Get the wrapped value.

        Returns:
            The value wrapped by this value object.

        Example:
            ```python
            class ProductName(ValueObject[str]):
                pass

            product = ProductName("iPhone 15")
            product.value  # 'iPhone 15'
            type(product.value)  # <class 'str'>
            ```
        """
        return self._value

    @override
    def __eq__(self, other: Self) -> bool:
        """
        Check equality with another value object of the same type.

        Two value objects are considered equal if their wrapped values are equal.

        Args:
            other: Another value object of the same type to compare with.

        Returns:
            True if both value objects have equal values, False otherwise.

        Example:
            ```python
            class UserId(ValueObject[int]):
                pass

            user1 = UserId(123)
            user2 = UserId(123)
            user3 = UserId(456)
            user1 == user2  # True
            user1 == user3  # False
            user1 == 123  # Different type, would raise error
            ```
        """
        if not isinstance(other, self.__class__):
            return False

        return self.value == other.value

    @override
    def __repr__(self) -> str:
        """
        Return a string representation suitable for debugging.

        Returns:
            A string in the format "ClassName(value)" that can be used to recreate the object.

        Example:
            ```python
            class OrderId(ValueObject[str]):
                pass

            order = OrderId("ORD-001")
            repr(order)  # "OrderId('ORD-001')"
            eval(repr(order))  # OrderId('ORD-001')
            ```
        """
        return f"{self.__class__.__name__}({self._value!r})"

    @override
    def __str__(self) -> str:
        """
        Return a human-readable string representation of the value.

        Returns:
            The string representation of the wrapped value.

        Example:
            ```python
            class Price(ValueObject[float]):
                pass

            price = Price(29.99)
            str(price)  # '29.99'
            print(f"Product costs ${price}")  # Product costs $29.99
            ```
        """
        return str(self._value)

    @override
    def __hash__(self) -> int:
        """
        Return the hash value of this value object.

        The hash is based on the wrapped value, making value objects suitable
        for use as dictionary keys or in sets.

        Returns:
            The hash value of the wrapped value.

        Example:
            ```python
            class CategoryId(ValueObject[str]):
                pass

            cat1 = CategoryId("electronics")
            cat2 = CategoryId("books")
            cat3 = CategoryId("electronics")

            categories = {cat1: "Electronics", cat2: "Books"}
            categories[cat3]  # 'Electronics'

            unique_categories = {cat1, cat2, cat3}
            len(unique_categories)  # 2
            ```
        """
        return hash(self._value)

    @override
    def __setattr__(self, name: str, value: T) -> None:
        """
        Prevent modification of the value after initialization.

        This method enforces immutability by raising an AttributeError for any
        attempt to modify the value or access non-existent attributes.

        Args:
            name: The name of the attribute being set.
            value: The value being assigned to the attribute.

        Raises:
            AttributeError: Always raised to prevent modification of the value object.

        Example:
            ```python
            class CustomerId(ValueObject[int]):
                pass

            customer = CustomerId(12345)
            customer._value = 54321  # Raises AttributeError

            customer.new_attribute = "test"  # Raises AttributeError
            ```
        """
        if name in self.__slots__:
            raise AttributeError("Cannot modify the value of a ValueObject")

        public_name = name.replace("_", "")
        public_slots = [slot.replace("_", "") for slot in self.__slots__]
        if public_name in public_slots:
            raise AttributeError("Cannot modify the value of a ValueObject")

        raise AttributeError(f"Class {self.__class__.__name__} object has no attribute '{name}'")



from sindripy.mothers.object_mother import ObjectMother


class BooleanPrimitivesMother(ObjectMother):
    """Generate boolean primitive values for testing."""

    @classmethod
    def any(cls) -> bool:
        """Generate any random boolean value."""
        return cls._faker().boolean()

    @staticmethod
    def true() -> bool:
        """Generate True value."""
        return True

    @staticmethod
    def false() -> bool:
        """Generate False value."""
        return False



"""Public facade for object mother helpers.

This module re-exports the available object mother implementations so
that projects using this library can import them from
``sindripy.mothers`` directly.
"""

from sindripy.mothers.identifiers.string_uuid_primitives_mother import StringUuidPrimitivesMother
from sindripy.mothers.object_mother import ObjectMother
from sindripy.mothers.primitives.boolean_primitives_mother import BooleanPrimitivesMother
from sindripy.mothers.primitives.float_primitives_mother import FloatPrimitivesMother
from sindripy.mothers.primitives.integer_primitives_mother import IntegerPrimitivesMother
from sindripy.mothers.primitives.list_primitives_mother import ListPrimitivesMother
from sindripy.mothers.primitives.string_primitives_mother import StringPrimitivesMother

__all__ = [
    "ObjectMother",
    "BooleanPrimitivesMother",
    "FloatPrimitivesMother",
    "IntegerPrimitivesMother",
    "ListPrimitivesMother",
    "StringPrimitivesMother",
    "StringUuidPrimitivesMother",
]



from sindripy.value_objects.decorators.validation import validate
from sindripy.value_objects.errors.incorrect_value_type_error import IncorrectValueTypeError
from sindripy.value_objects.errors.required_value_error import RequiredValueError
from sindripy.value_objects.value_object import ValueObject


class String(ValueObject[str]):
    """
    A value object that wraps string values with validation.

    This class provides a base implementation for creating value objects that
    represent string values in the domain. It ensures that the wrapped value
    is a valid string and not None.

    The class includes built-in validation for:
    - Required value (not None)
    - Type checking (must be a string)

    Inherits all functionality from ValueObject including immutability,
    equality comparison, string representation, and hashing.

    Example:
        ```python
        class Email(String):
            @validate
            def _validate_email_format(self, value: str) -> None:
                if "@" not in value:
                    raise ValueError("Invalid email format")

        email = Email("user@example.com")
        email.value  # 'user@example.com'
        ```
    """

    @validate
    def _ensure_has_value(self, value: str) -> None:
        if value is None:
            raise RequiredValueError

    @validate
    def _ensure_is_string(self, value: str) -> None:
        if not isinstance(value, str):
            raise IncorrectValueTypeError(value, str)



from collections.abc import Callable
from typing import Any, TypeVar

F = TypeVar("F", bound=Callable[..., Any])


def validate(func: F | None = None, *, order: int = 0) -> Callable[[F], F] | F:
    """Mark a method as a validator for ValueObject validation.

    Arguments:
        func: the function to decorate.
        order: order in which this validator should run relative to other validators in the same class. Lower numbers run first.
    """  # noqa: E501

    def wrapper(fn: F) -> F:
        if not isinstance(order, int):
            raise TypeError(f"Validation order {order} must be an integer. Got {type(order).__name__} type.")
        if order < 0:
            raise ValueError(f"Validation order {order} must be a positive value.")

        fn._is_validator = True
        fn._order = order
        return fn

    if func is not None:
        return wrapper(func)

    return wrapper



from sindripy.mothers.object_mother import ObjectMother


class StringUuidPrimitivesMother(ObjectMother):
    """Generate string UUID primitive values for testing."""

    @classmethod
    def any(cls) -> str:
        """Generate any random UUID string value."""
        return cls._faker().uuid4()

    @classmethod
    def invalid(cls) -> str:
        """Generate an invalid UUID string."""
        valid_uuid = cls.any()
        return valid_uuid[:-4]



from sindripy.value_objects.decorators.validation import validate
from sindripy.value_objects.errors.incorrect_value_type_error import IncorrectValueTypeError
from sindripy.value_objects.errors.required_value_error import RequiredValueError
from sindripy.value_objects.value_object import ValueObject


class Float(ValueObject[float]):
    """
    A value object that wraps float values with validation.

    This class provides a base implementation for creating value objects that
    represent float values in the domain. It ensures that the wrapped value
    is a valid float and not None. It accepts both positive and negative values.

    The class includes built-in validation for:
    - Required value (not None)
    - Type checking (must be a float)

    Inherits all functionality from ValueObject including immutability,
    equality comparison, string representation, and hashing.

    Example:
        ```python
        class Price(Float):
            @validate
            def _validate_positive(self, value: float) -> None:
                if value < 0:
                    raise ValueError("Price cannot be negative")

        price = Price(29.99)
        price.value  # 29.99
        str(price)  # '29.99'
        ```
    """

    @validate
    def _ensure_has_value(self, value: float) -> None:
        if value is None:
            raise RequiredValueError

    @validate
    def _ensure_value_is_float(self, value: float) -> None:
        if not isinstance(value, float):
            raise IncorrectValueTypeError(value, float)
